/* Code derived from:
  https://appusajeev.wordpress.com/2011/06/18/writing-a-linux-character-device-driver/
  and
  http://pete.akeo.ie/2011/08/writing-linux-device-driver-for-kernels.html
*/

#include <linux/module.h>
#include <linux/string.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <linux/pid.h>
#include <linux/cred.h>

#define dbg(format, arg...) do { if (debug) pr_info(CLASS_NAME ": %s: " format, __FUNCTION__, ## arg); } while (0)
#define err(format, arg...) pr_err(CLASS_NAME ": " format, ## arg)
#define info(format, arg...) pr_info(CLASS_NAME ": " format, ## arg)
#define warn(format, arg...) pr_warn(CLASS_NAME ": " format, ## arg)


#define DEVICE_NAME "yourtask"
#define CLASS_NAME "comp3000"

static struct class* yourtask_class = NULL;
static struct device* yourtask_device = NULL;
static int yourtask_major;
static pid_t thepid = -1;

static int yourtask_open(struct inode *the_inode, struct file *f)
{
        return 0;
}

static ssize_t yourtask_read(struct file *f, char *buf, size_t len, loff_t *offset)
{
        size_t i, msglen;
        pid_t task_pid;
        pid_t my_pid;
        struct task_struct *task;
        uid_t task_uid;
        uid_t task_euid;
        uid_t my_uid;
        uid_t my_euid;
        
        char message[100];
        
        if (*offset > 0) {
                return 0;
        }

        if (thepid == -1) {
                task_pid = task_tgid_vnr(current);
        } else {
                task_pid = thepid;
        }

        my_pid = task_tgid_vnr(current);
        my_uid = from_kuid_munged(current_user_ns(), current_uid());
        my_euid = from_kuid_munged(current_user_ns(), current_euid());

        task = pid_task(find_vpid(task_pid), PIDTYPE_PGID);
        task_uid = from_kuid_munged(current_user_ns(), task->cred->uid);
        task_euid = from_kuid_munged(current_user_ns(), task->cred->euid);
        
        if (my_uid == task_uid) {
                /* task is yours */
                if (task_euid == my_uid) {
                        /* is running as you as well */
                        snprintf(message, 100,
                                 "Task %d is yours!\n", task_pid);
                } else {
                        /* is running with another's privileges */
                        snprintf(message, 100,
                                 "Task %d is yours but acts as %d!\n",
                                 task_pid, task_euid);
                }
        } else {
                /* task is someone else's */
                if (task_euid == my_uid) {
                        /* but operating with our privileges */
                        snprintf(message, 100,
                                 "Task %d is %d's but acts as you!\n",
                                 task_pid, task_uid);
                } else if (task_euid == task_uid) {
                        /* acting with their privileges */
                        snprintf(message, 100,
                                 "Task %d is %d's!\n",
                                 task_pid, task_uid);
                } else {
                        /* acting with someone else's privileges */
                        snprintf(message, 100,
                                 "Task %d is %d's but acts as %d!\n",
                                 task_pid, task_uid, task_euid);
                }
        }
        
        msglen = strlen(message);

        if (len < msglen) {
                msglen = len;
        }

        for (i = 0; i < msglen; i++) {
                put_user(message[i], buf++);
        }

        *offset = i;

        return i;
}

static ssize_t yourtask_write(struct file *f, const char *buf, size_t len,
                              loff_t *offset)
{
        int p, error;
        ssize_t result;

        error = kstrtoint_from_user(buf, len, 10, &p);

        if (error == 0) {
                thepid = p;
        }

        *offset += len;
        result = (ssize_t) *offset;
        
        return result;
}

static int yourtask_release(struct inode *the_inode, struct file *f)
{
        printk(KERN_ALERT "Yourtask device closed\n");
        return 0;
}


static struct file_operations yourtask_fops = {
        .open = yourtask_open,
        .read = yourtask_read,
        .write = yourtask_write,
        .release = yourtask_release,
};


static char *yourtask_devnode(struct device *dev, umode_t *mode)
{
        if (mode)
	        *mode = 0666;
        return NULL;
}

static int __init yourtask_init(void)
{
        int retval;
  
        yourtask_major = register_chrdev(0, DEVICE_NAME, &yourtask_fops);
        if (yourtask_major < 0) {
                err("failed to register device: error %d\n", yourtask_major);
                retval = yourtask_major;
                goto failed_chrdevreg;
        }
 
        yourtask_class = class_create(THIS_MODULE, CLASS_NAME);
        if (IS_ERR(yourtask_class)) {
                err("failed to register device class '%s'\n", CLASS_NAME);
                retval = PTR_ERR(yourtask_class);
                goto failed_classreg;
        }
 
	yourtask_class->devnode = yourtask_devnode;

        yourtask_device = device_create(yourtask_class, NULL, MKDEV(yourtask_major, 0),
                                    NULL, DEVICE_NAME);

        if (IS_ERR(yourtask_device)) {
                err("failed to create device '%s'\n", DEVICE_NAME);
                retval = PTR_ERR(yourtask_device);
                goto failed_devreg;
        }
        
        info("Yourtask device registered using major %d.\n", yourtask_major);
        
        return 0;
        
 failed_devreg:
        class_unregister(yourtask_class);
        class_destroy(yourtask_class);
 failed_classreg:
        unregister_chrdev(yourtask_major, DEVICE_NAME);
 failed_chrdevreg:
        return -1;
}

static void __exit yourtask_exit(void)
{
        device_destroy(yourtask_class, MKDEV(yourtask_major, 0));
        class_unregister(yourtask_class);
        class_destroy(yourtask_class);
        unregister_chrdev(yourtask_major, "yourtask");
        info("Unloading Yourtask module.\n");
        return;
}

module_init(yourtask_init);
module_exit(yourtask_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Anil Somayaji <soma@scs.carleton.ca>");
MODULE_DESCRIPTION("Yourtask, give info about current task, character device module");
